//   ⨢ True-Cpp plugin. (c) Andrey Tretjakov,1999 v.1.0

#include "winmem.h"

#include <plugin.hpp>
#include <farkeys.hpp>

#include "eicoll.h"
#include "true-tpl.h"
#include "ttpl-lng.h"
#include "farintf.h"

#include "cregexp.h"

#include "syslog.h"

bool IsOldFar = true;

enum TExpandType
{
  expAnyWhere,
  expAtStart,
  expAtMiddle,
  expAtEnd,
  expOnBlank,
  expNone
};

#define MAX_REG_LEN 128
#define MAX_MAC_LEN 1024

struct TMacro
{
  char Keyword[MAX_REG_LEN];
  unsigned MinLen;
  int FARKey;
  TExpandType ExpandType;
  char MacroText[MAX_MAC_LEN];
};

struct TIndent
{
  char mask[MAX_REG_LEN], relative[MAX_REG_LEN], immChar;
  int indent[2], start, BracketsMode, bracketLink;
};

struct TBracket
{
  char open[MAX_REG_LEN], close[MAX_REG_LEN];
};

struct TComment
{
  char mask[MAX_REG_LEN];
};

struct TLang
{
  char mask[MAX_REG_LEN], desc[MAX_REG_LEN], imm[MAX_REG_LEN];
  int ignoreCase, TabSize;
  TCollection macroColl, indentColl, bracketColl, commentColl;
};

static void delLang(void *it)
{
  TLang *lng = (TLang*)it;
  lng->macroColl.removeAll();
  lng->indentColl.removeAll();
  lng->bracketColl.removeAll();
  lng->commentColl.removeAll();
}

class TLangCollection: public TCollection
{
  public:
  TLangCollection() : TCollection(5, 5, delLang) {};
};

static TLangCollection langColl;
static int pluginBusy = 0;
static int pluginStop = 0;

static char *addExt(char *path, char *defExt)
{
  int n = lstrlen(path)-1;
  for ( int i = n ; i >= 0 ; i-- )
  {
    if ( path[i] == '.' )
    {
      lstrcpy(path+i, defExt);
      break;
    }
    if ( ( path[i] == '\\' ) || ( path[i] == ':' ) || ( i == 0 ) )
      return lstrcat(path, defExt);
  }
  return path;
}

static char *lstrchr(char *s, int c)
{
  char *p;
  for ( p = s ; *p ; p++ )
    if ( *p == c )
      return p;
  return c ? NULL : p;
}

static inline int isCharSpace(char c)
{
  return lstrchr(" \t\r\n", c) != NULL;
}

static char *skipSpaces(char*& line)
{
  while ( *line && isCharSpace(*line) )
    line++;
  return line;
}

static char *getWord(char*& line, char *kwd)
{
  char *k = kwd;
  if ( *skipSpaces(line) == '\"')
  {
    while ( *++line )
    {
      if ( *line == '\\' && line[1] == '\"' )
        *k++ = *++line;
      else if ( *line == '\"' )
      {
        if ( *skipSpaces(++line) != '\"' )
          break;
      }
      else
        *k++ = *line;
    }
    *k = 0;
  }
  else
  {
    while ( *line && !isCharSpace(*line) )
      *k++ = *line++;
    *k = 0;
  }
  return kwd;
}

static char *parseItem(char*& line, char *kwd, char *value)
{
  if ( IsCharAlpha(*skipSpaces(line)) )
  {
    char *k = kwd;
    while ( *line && IsCharAlpha(*line) )
      *k++ = *line++;
    *k = *value = 0;
    if ( *skipSpaces(line) == '=' )
    {
      line++;
      getWord(line, value);
    }
    return line;
  }
  return NULL;
}

static char *getItem(char*& line, char *kwd)
{
  if ( *skipSpaces(line) == '<')
  {
    char *k = kwd;
    line++;
    skipSpaces(line);
    while ( *line && IsCharAlpha(*line) )
      *k++ = *line++;
    *k = 0;
    k = line;
    int inQuote = 0;
    while ( *++line )
    {
      if ( inQuote )
      {
        if ( *line == '\\' && line[1] == '\"' )
          line++;
        else if ( *line == '\"' )
          inQuote = 0;
      }
      else
      {
        if ( *line == '>' )
        {
          line++;
          break;
        }
        else if ( *line == '\"' )
          inQuote = 1;
      }
    }
    return k;
  }
  return NULL;
}

static void InitMacro(char *file)
{
  HANDLE hScreen=Info.SaveScreen(0,0,-1,-1);
  const char *MsgItems[]={GetMsg(MMenuString),GetMsg(MLoading)};
  Info.Message(Info.ModuleNumber,0,NULL,MsgItems,sizeof(MsgItems)/sizeof(MsgItems[0]),0);
  HANDLE s = CreateFile(file, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
  if ( s != INVALID_HANDLE_VALUE )
  {
    unsigned fileSize = GetFileSize(s, NULL);
    if ( fileSize )
    {
      char *fileBuff = new char[fileSize+1];
      if ( fileBuff )
      {
        DWORD rb = 0;
        if ( ReadFile(s, fileBuff, fileSize, &rb, NULL) && rb )
        {
          fileBuff[rb] = 0;
          char *item, *p = fileBuff, name[32], value[1024];
          TLang *lng = NULL;
          TCollection dummyLang;
          memcpy(&langColl, &dummyLang, sizeof(TLangCollection));
          while ( ( item = getItem(p, name) ) != NULL )
          {
            if ( !stricmp(name, "Language") )
            {
              lng = new TLang;
              lng->ignoreCase = lng->TabSize = 0;
              lng->mask[0] = lng->desc[0] = 0;
              getWord(item, lng->mask);
              lng->imm[0] = 0;
              while ( parseItem(item, name, value) )
              {
                if ( !stricmp(name, "Desc") )
                  lstrcpy(lng->desc, value);
                if ( !stricmp(name, "IgnoreCase") )
                  lng->ignoreCase = FSF.atoi(value);
                if ( !stricmp(name, "Tab") )
                  lng->TabSize = FSF.atoi(value);
              }
              langColl.insert(lng);
            }
            else if ( lng != NULL )
            {
              if ( !stricmp(name, "Expand") )
              {
                TMacro *tmpm = new TMacro;
                tmpm->ExpandType = expOnBlank;
                tmpm->Keyword[0] = tmpm->MacroText[0] = 0;
                tmpm->FARKey = -1;
                getWord(item, tmpm->Keyword);
                int validTag = true;
                if ( tmpm->Keyword[0] == 0 )
                  validTag = false;
                char *opDelimiter = lstrchr(tmpm->Keyword, '|');
                if ( opDelimiter )
                {
                  *opDelimiter = 0;
                  tmpm->MinLen = lstrlen(tmpm->Keyword);
                  lstrcpy(opDelimiter, opDelimiter+1);
                }
                else
                  tmpm->MinLen = lstrlen(tmpm->Keyword);
                while ( parseItem(item, name, value) )
                {
                  if      ( !stricmp(name, "To") )
                    lstrcpy(tmpm->MacroText, value);
                  else if ( !stricmp(name, "Key") )
                  {
                    validTag = true;
                    tmpm->FARKey = FSF.FarNameToKey(value);
                  }
                  else if ( !stricmp(name, "At") )
                  {
                    if      ( !stricmp(value, "AnyWhere") ) tmpm->ExpandType = expAnyWhere;
                    else if ( !stricmp(value, "Start" ) )   tmpm->ExpandType = expAtStart;
                    else if ( !stricmp(value, "Middle") )   tmpm->ExpandType = expAtMiddle;
                    else if ( !stricmp(value, "End") )      tmpm->ExpandType = expAtEnd;
                    else if ( !stricmp(value, "Blank") )    tmpm->ExpandType = expOnBlank;
                  }
                }
                if ( validTag )
                  lng->macroColl.insert(tmpm);
              }
              else if ( !stricmp(name, "Comment") )
              {
                TComment *tmpc = new TComment;
                tmpc->mask[0] = 0;
                getWord(item, tmpc->mask);
                lng->commentColl.insert(tmpc);
              }
              else if ( !stricmp(name, "Indent") )
              {
                TIndent *tmpi = new TIndent;
                tmpi->mask[0] = tmpi->relative[0] = 0;
                tmpi->BracketsMode = tmpi->start = tmpi->immChar = 0;
                getWord(item, tmpi->mask);
                for ( int i = 0 ; i < 2 ; i++ )
                  tmpi->indent[i] = 0xFFFF;
                if ( tmpi->mask[0] )
                {
                  while ( parseItem(item, name, value) )
                  {
                    if      ( !stricmp(name, "Line") )     tmpi->indent[0] = FSF.atoi(value);
                    else if ( !stricmp(name, "Next") )     tmpi->indent[1] = FSF.atoi(value);
                    else if ( !stricmp(name, "Imm") )      tmpi->immChar = *value;
                    else if ( !stricmp(name, "Start") )    tmpi->start = FSF.atoi(value);
                    else if ( !stricmp(name, "Relative") ) lstrcpy(tmpi->relative, value);
                  }
                  if ( tmpi->immChar )
                  {
                    if ( !lstrchr(lng->imm, tmpi->immChar) )
                    {
                      int len = lstrlen(lng->imm);
                      lng->imm[len++] = tmpi->immChar;
                      lng->imm[len] = 0;
                    }
                  }
                }
                lng->indentColl.insert(tmpi);
              }
              else if ( !stricmp(name, "Bracket") )
              {
                TBracket *tmpb = new TBracket;
                TIndent  *tmp0 = new TIndent;
                TIndent  *tmp1 = new TIndent;
                tmpb->open[0] = tmpb->close[0] = 0;
                tmp0->relative[0] = 0;
                tmp0->start = tmp0->immChar = 0;
                tmp0->BracketsMode = 0;
                tmp1->relative[0] = 0;
                tmp1->start = tmp1->immChar = 0;
                tmp1->BracketsMode = 1;
                tmp1->bracketLink = lng->bracketColl.getCount();
                getWord(item, tmpb->open);
                if ( parseItem(item, name, tmpb->close) )
                {
                  if ( !stricmp(name, "Match") )
                  {
                    lstrcpy(tmp0->mask, tmpb->open);
                    lstrcpy(tmp1->mask, tmpb->close);
                    tmp0->indent[0] = 0xFFFF;
                    tmp0->indent[1] = 1;
                    tmp1->indent[0] = 0;
                    tmp1->indent[1] = 0xFFFF;
                    if ( tmpb->open[0] && tmpb->close[0] )
                    {
                      while ( parseItem(item, name, value) )
                      {
                        if      ( !stricmp(name, "Line") )     tmp0->indent[0] = FSF.atoi(value);
                        else if ( !stricmp(name, "Shift") )    tmp0->indent[1] = FSF.atoi(value);
                        else if ( !stricmp(name, "Next") )     tmp1->indent[1] = FSF.atoi(value);
                        else if ( !stricmp(name, "ImmOpen") )  tmp0->immChar = *value;
                        else if ( !stricmp(name, "ImmClose") ) tmp1->immChar = *value;
                        else if ( !stricmp(name, "Start") )    tmp0->start = FSF.atoi(value);
                      }
                      if ( tmp0->immChar )
                      {
                        if ( !lstrchr(lng->imm, tmp0->immChar) )
                        {
                          int len = lstrlen(lng->imm);
                          lng->imm[len++] = tmp0->immChar;
                          lng->imm[len] = 0;
                        }
                      }
                      if ( tmp1->immChar )
                      {
                        if ( !lstrchr(lng->imm, tmp1->immChar) )
                        {
                          int len = lstrlen(lng->imm);
                          lng->imm[len++] = tmp1->immChar;
                          lng->imm[len] = 0;
                        }
                      }
                    }
                    lng->bracketColl.insert(tmpb);
                    lng->indentColl.insert(tmp0);
                    lng->indentColl.insert(tmp1);
                  }
                }
              }
            }
          }
        }
        delete [] fileBuff;
      }
    }
  }
  CloseHandle(s);
  Info.RestoreScreen(hScreen);
}

static void DoneMacro()
{
  langColl.removeAll();
}

static inline char *StrLower(char *str)
{
  FSF.LLowerBuf(str, lstrlen(str));
  return str;
}

static TMacro *CheckMacroPos(TMacro *mc, TExpandType exp)
{
  if ( exp == expAnyWhere )
    return mc;
  TMacro *rmc = NULL;
  switch ( mc->ExpandType )
  {
    case expAnyWhere:
      rmc = mc;
      break;
    case expAtMiddle:
      if ( exp == expAtMiddle )
        rmc = mc;
      break;
    case expOnBlank:
      if ( exp == expOnBlank )
        rmc = mc;
      break;
    case expAtStart:
      if ( exp == expOnBlank || exp == expAtStart )
        rmc = mc;
      break;
    case expAtEnd:
      if ( exp == expOnBlank || exp == expAtEnd )
        rmc = mc;
      break;
    default:
      break;
  }
  return rmc;
}

static TMacro *FindMacro(TLang *lng, char *line, unsigned l, TExpandType exp)
{
  for ( unsigned i = 0 ; i < lng->macroColl.getCount() ; i++ )
  {
    TMacro *mm = (TMacro*)(lng->macroColl.at(i));
    if ( l >= mm->MinLen && l <= (unsigned)lstrlen(mm->Keyword) )
    {
      TMacro *rmc = CheckMacroPos(mm, exp);
      if ( rmc )
      {
        if ( lng->ignoreCase )
        {
          if ( !FSF.LStrnicmp(mm->Keyword, line, l) )
            return rmc;
        }
        else
        {
          if ( !strncmp(mm->Keyword, line, l) )
            return rmc;
        }
      }
    }
  }
  return NULL;
}

TExpandType GetMacroPos(char *line, char *ptr, int len)
{
  char *startOfLine = line, *endOfLine = lstrchr(line, 0);
  while ( isCharSpace(*startOfLine) )
    startOfLine++;
  endOfLine -= len;
  TExpandType exp = expNone;
  if      ( startOfLine == endOfLine )             exp = expOnBlank;
  else if ( ptr > startOfLine && ptr < endOfLine ) exp = expAtMiddle;
  else if ( ptr == startOfLine )                   exp = expAtStart;
  else if ( ptr == endOfLine )                     exp = expAtEnd;
  return exp;
}

static TMacro *FindMacroKey(TLang *lng, const INPUT_RECORD *Rec)
{
  int rKey = FSF.FarInputRecordToKey(Rec);
  for ( unsigned i = 0 ; i < lng->macroColl.getCount() ; i++ )
  {
    TMacro *mm = (TMacro*)(lng->macroColl.at(i));
    if ( mm->FARKey == rKey )
      return mm;
  }
  return NULL;
}

static TMacro *FindMacroKeyEx(TLang *lng, TMacro *mc, TExpandType at)
{
  if ( mc->FARKey != -1 )
    for ( unsigned i = 0 ; i < lng->macroColl.getCount() ; i++ )
    {
     TMacro *mm = (TMacro*)(lng->macroColl.at(i));
     if ( ( mc->FARKey == mm->FARKey ) && ( CheckMacroPos(mm, at) != NULL ) )
       return mm;
    }
  return NULL;
}

static char userString[10][80] = { "", "", "", "", "", "", "", "", "", "" };
static char titleString[80] = "", *userStr[10], *titleStr;

static void nullUserString(void)
{
  for ( int i = 0 ; i < 10 ; i++ )
    userStr[i] = NULL;
}

static int insertUserString(int n)
{
  if ( userStr[n] == NULL )
  {
    struct EditorInfo ei;
    static char *h = "TrueTpl";
    Info.EditorControl(ECTL_GETINFO, &ei);
    struct InitDialogItem Init[] =
    {
      { DI_DOUBLEBOX, 1, 0, 22, 3, 0,        0,               titleString    },
      { DI_EDIT,      3, 1, 20, 0, (DWORD)h, DIF_HISTORY,     ""             },
      { DI_BUTTON,    0, 2,  0, 0, 0,        DIF_CENTERGROUP, (char*)MOK     },
      { DI_BUTTON,    0, 2,  0, 0, 0,        DIF_CENTERGROUP, (char*)MCancel }
    };
    int CoorX = ei.CurPos-ei.LeftPos;
    int CoorY = ei.CurLine-ei.TopScreenLine;
    int mx = ( CoorX+28 > ei.WindowSizeX ) ? CoorX-32 : CoorX;
    int my = ( CoorY+3  > ei.WindowSizeY ) ? CoorY-2  : CoorY;
    int size = sizeof(Init)/sizeof(Init[0]);
    FarDialogItem *Item = new FarDialogItem[size];
    InitDialogItems(Init, Item, size);
    lstrcpy(Item[1].Data, userString[n]);
    Item[1].Focus = Item[2].DefaultButton = 1;
    int Result = Info.Dialog(Info.ModuleNumber,mx,my,mx+23,my+3,NULL,Item,size);
    if ( Result == 2 )
      userStr[n] = lstrcpy(userString[n], Item[1].Data);
    delete [] Item;
  }
  if ( userStr[n] )
    Info.EditorControl(ECTL_INSERTTEXT, userStr[n]);
  return userStr[n] != NULL;
}

struct TEditorPos
{
  TEditorPos() { Default(); }
  int Row, Col;
  int TopRow,LeftCol;
  void Default(void) { Row = Col = TopRow = LeftCol = -1; }
};

static TEditorPos EditorGetPos(void)
{
  TEditorPos r;
  EditorInfo ei;
  Info.EditorControl(ECTL_GETINFO, &ei);
  r.Row = ei.CurLine;
  r.Col = ei.CurPos;
  r.TopRow = ei.TopScreenLine;
  r.LeftCol = ei.LeftPos;
  return r;
}

static void EditorSetPos(TEditorPos pos)
{
  EditorSetPosition sp;
  sp.CurLine = pos.Row;
  sp.CurPos = pos.Col;
  sp.TopScreenLine = pos.TopRow;
  sp.LeftPos = pos.LeftCol;
  sp.CurTabPos = sp.Overtype = -1;
  Info.EditorControl(ECTL_SETPOSITION,&sp);
}

void EditorSetPos(int line, int col, int topline = -1, int leftcol = -1)
{
  EditorSetPosition sp;
  sp.CurLine = line;
  sp.CurPos = col;
  sp.TopScreenLine = topline;
  sp.LeftPos = leftcol;
  sp.CurTabPos = sp.Overtype = -1;
  Info.EditorControl(ECTL_SETPOSITION, &sp);
}

static void EditorProcessKey(char ascii)
{
  INPUT_RECORD tr;
  tr.EventType = KEY_EVENT;
  tr.Event.KeyEvent.bKeyDown = true;
  tr.Event.KeyEvent.wRepeatCount = 1;
  tr.Event.KeyEvent.wVirtualKeyCode = 0;
  tr.Event.KeyEvent.uChar.AsciiChar = ascii;
  tr.Event.KeyEvent.dwControlKeyState = 0;
  Info.EditorControl(ECTL_PROCESSINPUT, &tr);
}

static inline void EditorProcessFARKey(int key)
{
  Info.EditorControl(ECTL_PROCESSKEY, (void*)key);
}

static inline void EditorInsertText(char *s)
{
  Info.EditorControl(ECTL_INSERTTEXT, (void*)s);
}

static inline void redraw(void)
{
  Info.EditorControl(ECTL_REDRAW, 0);
}

static void EditorProcessTabKey(int TabSize, int pos, int n)
{
  if ( n > 0 )
  {
    int idnt;
    char symbol;
    if ( TabSize ) // if TabSize = 0 use Tabs! [updated by DO]
    {
      idnt = n*TabSize-pos%TabSize;
      symbol = ' ';
    }
    else
    {
      idnt = n;
      symbol = '\t';
    }
    if ( idnt > 0 )
    {
      char s[1024];
      memset(s, symbol, idnt);
      s[idnt] = 0;
      EditorInsertText(s);
      redraw();
    }
  }
}

static void EditorProcessReturnKey(int before = -1, int after = -1)
{
  INPUT_RECORD tr;
  tr.EventType = KEY_EVENT;
  tr.Event.KeyEvent.bKeyDown = true;
  tr.Event.KeyEvent.wRepeatCount = 1;
  tr.Event.KeyEvent.wVirtualKeyCode = VK_RETURN;
  tr.Event.KeyEvent.dwControlKeyState = 0;
  if ( before != -1 )
    pluginBusy = before;
  Info.EditorControl(ECTL_PROCESSINPUT, &tr);
  if ( after != -1 )
    pluginBusy = after;
}

static void EditorGetString(EditorGetString *gs, int line = -1)
{
  gs->StringNumber = line;
  Info.EditorControl(ECTL_GETSTRING, gs);
}

static void EditorSetString(char *src, int line = -1)
{
  struct EditorSetString st;
  st.StringNumber = line;
  st.StringText = src;
  st.StringEOL = 0;
  st.StringLength = lstrlen(src);
  Info.EditorControl(ECTL_SETSTRING, &st);
}

static void EditorGoToHome(void)
{
  struct EditorGetString gs;
  EditorProcessFARKey(KEY_HOME);
  EditorGetString(&gs);
  if ( isCharSpace(gs.StringText[0]) )
    EditorProcessFARKey(KEY_RIGHT|KEY_CTRL);
}

static inline void insIns(char *ins, int& insPos)
{
  if ( *ins )
  {
    EditorInsertText(ins);
    insPos = *ins = 0;
    redraw();
  }
}

static void RunMacro(TMacro *m, TLang *lng)
{
  int insPos = 0, setPos = 0;
  char ins[1024] = "";
  TEditorPos pos, curPos;
  pluginBusy = 1;
  nullUserString();
  for ( char *p = m->MacroText ; *p ; p++ )
  {
    char *cb;
    if ( *p == '\\' )
    {
      insIns(ins, insPos);
      switch ( *++p )
      {
        case 'p':
          setPos = 1;
          pos = EditorGetPos();
          break;
        case 0:
        case '\\':
          EditorProcessFARKey(KEY_BACKSLASH);
          break;
        case 't':
          curPos = EditorGetPos();
          EditorProcessTabKey(lng->TabSize, curPos.Col, 1);
          break;
        case 'h':
          EditorGoToHome();
          break;
        case 'b':
          EditorProcessFARKey(KEY_BS);
          break;
        case '^':
          EditorProcessFARKey(KEY_UP);
          break;
        case 'v':
          EditorProcessFARKey(KEY_DOWN);
          break;
        case '<':
          EditorProcessFARKey(KEY_LEFT);
          break;
        case '>':
          EditorProcessFARKey(KEY_RIGHT);
          break;
        case '[':
          EditorProcessFARKey(KEY_HOME);
          break;
        case ']':
          EditorProcessFARKey(KEY_END);
          break;
        case '(':
          EditorProcessFARKey(KEY_LEFT|KEY_CTRL);
          break;
        case ')':
          EditorProcessFARKey(KEY_RIGHT|KEY_CTRL);
          break;
        case 'r':
          EditorProcessReturnKey(0, 1);
          break;
        case 'n':
          EditorProcessReturnKey();
          break;
        case '!':
          pluginBusy = 0;
          EditorProcessKey(*++p);
          pluginBusy = 1;
        case '?':
          if ( p[1] == '\'' )
          {
            p++;
            *(titleStr = titleString) = 0;
            while ( *++p && *p != '\'' )
              *titleStr++ = *p;
            *titleStr = 0;
          }
          break;
        case '0': case '1': case '2': case '3': case '4':
        case '5': case '6': case '7': case '8': case '9':
          if ( p[1] == '\'' )
          {
            p++;
            *(titleStr = userString[*p-'0']) = 0;
            while ( *++p && *p != '\'' )
              *titleStr++ = *p;
            *titleStr = 0;
          }
          if ( !insertUserString(*p-'0') )
            userStr[*p-'0'] = lstrcpy(userString[*p-'0'], "");
          break;
        case '{':
          if ( ( cb = lstrchr(p+1, '}') ) == NULL )
            EditorProcessKey(*p);
          else
          {
            char ch = *cb;
            *cb = 0;
            EditorProcessFARKey(FSF.FarNameToKey(p+1));
            *cb = ch;
            p = cb;
          }
          break;
        default:
          pluginBusy = 0;
          EditorProcessKey(*p);
          pluginBusy = 1;
          break;
      }
    }
    else
    {
      ins[insPos++] = *p;
      ins[insPos] = 0;
    }
  }
  insIns(ins, insPos);
  pluginBusy = 0;
  if ( setPos )
    EditorSetPos(pos);
}

int WINAPI _export GetMinFarVersion(void)
{
  // return FARMANAGERVERSION;
  return 0x01410146UL;
}

HANDLE heap;

void WINAPI _export SetStartupInfo(const struct PluginStartupInfo *Info)
{
  initMem();
  ::Info = *Info;
  IsOldFar = true;
  if ( Info->StructSize >= sizeof(struct PluginStartupInfo) )
  {
    ::FSF = *Info->FSF;
    ::Info.FSF = &::FSF;
    IsOldFar = false;
    char macfile[NM];
    FSF.sprintf(PluginRootKey,"%s\\TRUE-TPL",Info->RootKey);
    GetRegKey(HKEY_CURRENT_USER, "", "Disable", pluginStop, 0);
    addExt(lstrcpy(macfile, Info->ModuleName), ".ini");
    InitMacro(macfile);
  }
}

TEICollection *eList = NULL;

static void initEList(void)
{
  if ( eList == NULL )
    eList = new TEICollection;
}

void WINAPI _export ExitFAR()
{
  DoneMacro();
  if ( eList )
  {
    eList->done();
    delete eList;
  }
}

void WINAPI _export GetPluginInfo(struct PluginInfo *Info)
{
  if ( !IsOldFar )
  {
    static char *PluginMenuStrings[1];
    Info->StructSize = sizeof(*Info);
    Info->Flags = PF_EDITOR|PF_DISABLEPANELS;
    Info->DiskMenuStringsNumber = 0;
    Info->PluginConfigStringsNumber = 0;
    PluginMenuStrings[0] = GetMsg(MMenuString);
    Info->PluginMenuStrings = PluginMenuStrings;
    Info->PluginMenuStringsNumber = 1;
    static char *PluginCfgStrings[1];
    PluginCfgStrings[0]=GetMsg(MSetupString);
    Info->PluginConfigStrings=PluginCfgStrings;
    Info->PluginConfigStringsNumber=sizeof(PluginCfgStrings)/sizeof(PluginCfgStrings[0]);
  }
}

int Config(void)
{
  struct InitDialogItem InitItems[] =
  {
    { DI_DOUBLEBOX, 3,1,33,  5,0,         0,               (char*)MSetupTitle },
    { DI_CHECKBOX,  5,2, 0,  0,pluginStop,0,               (char*)MDisable    },
    { DI_TEXT,      5,3, 0,255,0,         0,               NULL               },
    { DI_BUTTON,    0,4, 0,  0,0,         DIF_CENTERGROUP, (char*)MReload     },
    { DI_BUTTON,    0,4, 0,  0,0,         DIF_CENTERGROUP, (char*)MCancel     }
  };
  int ExitCode = ExecDialog(InitItems,sizeof(InitItems)/sizeof(InitItems[0]),1,3,NULL);
  if ( ExitCode == 3 )
  {
    SetRegKey(HKEY_CURRENT_USER, "", "Disable", pluginStop = InitItems[1].Selected);
    DoneMacro();
    char macfile[260];
    addExt(lstrcpy(macfile, Info.ModuleName), ".ini");
    InitMacro(macfile);
    return true;
  }
  return false;
}

int WINAPI _export Configure(int ItemNumber)
{
  if ( ItemNumber )
    return false;
  return Config();
}

static inline int matched(char *mask, char *value)
{
  return FSF.ProcessName(mask, value, PN_CMPNAMELIST|PN_SKIPPATH);
}

static char *FirstNonSpace(char *s)
{
  char *i = s;
  while ( *i )
  {
    if ( !isCharSpace(*i) )
      return i;
    i++;
  }
  return NULL;
}

static int strMatch(char *s, char *pattern, char *prefix, char *suffix, int *l, int n)
{
  int ret = false;
  char slashPattern[320];
  static CRegExp reg;
  SMatches m;
  lstrcat(lstrcat(lstrcpy(slashPattern, prefix), pattern), suffix);
  if ( reg.SetExpr(slashPattern) )
  {
    if ( slashPattern[1] == '^' )
      reg.SetNoMoves(true);
    ret = reg.Parse(s, &m) ? true : false;
    if ( ret && l )
    {
      l[0] = m.s[n];
      l[1] = m.e[n];
    }
  }
  return ret;
}

static char *LookForOpenBracket(TEditorPos& p, int ignoreCase, TLang *lng, char* openBr)
{
  static char *c, buff[1024];
  static struct EditorGetString gs;
  int ret = -1;
  while ( --p.Row >= 0 )
  {
    EditorSetPos(p);
    EditorGetString(&gs);
    if ( gs.StringLength )
    {
      if ( ( c = FirstNonSpace(gs.StringText) ) != NULL )
      {
        if ( strMatch(lstrcpy(buff, FSF.Trim(c)), openBr, "/^", ignoreCase ? "$/i" : "$/", NULL, 0) )
        {
          ret = c-gs.StringText;
          break;
        }
        else
        {
          for ( unsigned i = 0 ; i < lng->bracketColl.getCount() ; i++ )
          {
            TBracket *br = (TBracket*)(lng->bracketColl.at(i));
            if ( strMatch(buff, br->close, "/^", ignoreCase ? "$/i" : "$/", NULL, 0 ) )
            {
              if ( LookForOpenBracket(p, ignoreCase, lng, br->open) == NULL )
                return NULL;
              break;
            }
          }
        }
      }
    }
  }
  if (ret != -1){
    memcpy(buff, gs.StringText, ret);
    buff[ret] = 0;
    return buff;
  } else return NULL;
}

static char *FindOpenBracket(int ignoreCase, TLang *lng, char *openBr)
{
  TEditorPos p, o = EditorGetPos();
  p.Default();
  p.Row = o.Row;
  char *ret = LookForOpenBracket(p, ignoreCase, lng, openBr);
  EditorSetPos(o);
  return ret;
}

static char *prevStringIndent(void)
{
  static char buff[512];
  struct EditorGetString gs;
  int ret = 0;
  TEditorPos p, o = EditorGetPos();
  p.Default();
  p.Row = o.Row;
  while ( --p.Row >= 0 )
  {
    EditorSetPos(p);
    EditorGetString(&gs);
    if ( gs.StringLength )
    {
      char *c = FirstNonSpace(gs.StringText);
      if ( c )
      {
        ret = c-gs.StringText;
        break;
      }
    }
  }
  EditorSetPos(o);
  memcpy(buff,gs.StringText,ret);
  buff[ret]=0;
  return buff;
}

static int shiftLine(int i, int TabSize)
{
  TEditorPos pos = EditorGetPos();
  pluginBusy = 1;
  if ( i > 0 )
    EditorProcessTabKey(TabSize, pos.Col, i);
  else if ( i < 0 )
  {
    int ind = -i*TabSize;
    for ( int j = 0 ; j < ind ; j++ )
      EditorProcessFARKey(KEY_BS);
  }
  TEditorPos pos2 = EditorGetPos();
  pluginBusy = 0;
  return pos2.Col-pos.Col;
}

static int TryIndent(int ret, TLang *lng, int j, char *str, char *realStr, int oldIndent)
{
  char line[1024];
  TIndent *tmpi = (TIndent*)(lng->indentColl.at(j));
  lstrcpy(line, tmpi->mask);
  int ic = lng->ignoreCase;
  if ( ic )
    StrLower(line);
  int bounds[2];
  if ( strMatch(str, line, "/^", ic ? "$/i" : "$/", bounds, tmpi->start) )
  {
    int iCurr = tmpi->indent[0];
    int iNext = tmpi->indent[1];
    if ( iCurr != 0xFFFF )
    {
      char *opencol = NULL;
      TEditorPos pos = EditorGetPos();
      if ( *(tmpi->relative) )
        opencol = FindOpenBracket(ic, lng, tmpi->relative);
      else if ( tmpi->BracketsMode )
        opencol = FindOpenBracket(ic, lng, ((TBracket*)(lng->bracketColl.at(tmpi->bracketLink)))->open);
      if ( opencol == NULL )
        opencol = prevStringIndent();
      size_t olen = lstrlen(opencol);
      char *ptr = new char[olen+lstrlen(str)+1];
      lstrcat(lstrcpy(ptr, opencol), realStr);
      EditorSetString(ptr);
      EditorSetPos(-1, olen);
      EditorSetPos(-1, pos.Col+olen-oldIndent+shiftLine(iCurr, lng->TabSize));
    }
    if ( ret )
    {
      EditorProcessReturnKey(1, 0);
      if ( tmpi->start )
      {
        pluginBusy = 1;
        for ( int k = bounds[0] ; k < bounds[1] ; k++ )
          EditorProcessKey(realStr[k]);
        pluginBusy = 0;
      }
      if ( iNext != 0xFFFF )
        shiftLine(iNext, lng->TabSize);
    }
    else
      redraw();
    return 1;
  }
  return 0;
}

static int inComment(TLang *lng, char *realLine, unsigned pos)
{
  int ic = lng->ignoreCase;
  for ( unsigned i = 0 ; i < lng->commentColl.getCount() ; i++ )
  {
    char line[1024];
    TComment *tmpc = (TComment*)(lng->commentColl.at(i));
    lstrcpy(line, realLine);
    if ( ic )
      StrLower(line);
    int bounds[2];
    if ( strMatch(line, tmpc->mask, "/", ic ? "/i" : "/", bounds, 0) )
      if ( pos >= (unsigned)bounds[0] && pos <= (unsigned)bounds[1] )
        return 1;
  }
  return 0;
}

static bool checkMultipleChoice(TLang *lng, TMacro*&mc, TExpandType exp)
{
  const char *menuPrefix = "\\~";
  unsigned pl = lstrlen(menuPrefix);
  if ( (unsigned)lstrlen(mc->MacroText) > pl && !strncmp(mc->MacroText, menuPrefix, pl) )
  {
    unsigned lc = 0;
    char *p = mc->MacroText, *p1;
    while ( ( p1 = strstr(p, menuPrefix) ) != NULL )
    {
      p = p1+pl;
      lc++;
    }
    if ( lc )
    {
      FarMenuItem *amenu = new FarMenuItem[lc];
      TMacro **amacro = new TMacro*[lc];
      if ( amenu )
      {
        unsigned i = 0;
        p = mc->MacroText;
        while ( ( p1 = strstr(p, menuPrefix) ) != NULL )
        {
          p = p1+pl;
          char ch, *p2 = strstr(p, menuPrefix);
          if ( p2 )
          {
            ch = *p2;
            *p2 = 0;
          }
          TMacro *tm = NULL;
          char *p3 = lstrchr(p, '=');
          if ( p3 )
          {
            *p3 = 0;
            tm = FindMacro(lng, p, lstrlen(p), exp);
            lstrcpy(amenu[i].Text, p3+1);
            *p3 = '=';
          }
          else
          {
            tm = FindMacro(lng, p, lstrlen(p), exp);
            lstrcpy(amenu[i].Text, p);
          }
          if ( p2 )
            *p2 = ch;
          if ( tm )
          {
            amacro[i] = tm;
            amenu[i].Selected = amenu[i].Checked = amenu[i].Separator = false;
            i++;
          }
        }
        int res = Info.Menu(Info.ModuleNumber,-1,-1,0,FMENU_WRAPMODE,GetMsg(MSelectMacro),NULL,NULL,NULL,NULL,amenu,i);
        bool retval = false;
        if ( res != -1 && amacro[res] )
        {
          mc = amacro[res];
          retval = true;
        }
        delete [] amenu;
        delete [] amacro;
        return retval;
      }
    }
  }
  return true;
}

static int TryMacro(TLang *lng, struct EditorGetString& gs, TEditorPos& ep, char *line)
{
  if ( !isCharSpace(gs.StringText[ep.Col-1]) )
  {
    char *ptr = lstrcpy(line, gs.StringText)+ep.Col;
    char ch = *ptr;
    *ptr-- = 0;
    while ( (ptr>=line) && !isCharSpace(*ptr) )
      ptr--;
    int len = lstrlen(++ptr);
    line[ep.Col] = ch;
    if ( len )
    {
      if ( !inComment(lng, line, ep.Col) )
      {
        TExpandType at = GetMacroPos(line, ptr, len);
        TMacro *mc = FindMacro(lng, ptr, len, at);
        if ( mc && checkMultipleChoice(lng, mc, at) )
        {
          pluginBusy = 1;
          for ( int i = 0 ; i < len ; i++ )
            EditorProcessFARKey(KEY_BS);
          pluginBusy = 0;
          RunMacro(mc, lng);
          redraw();
          return 1;
        }
      }
    }
  }
  return 0;
}

static int TryMacroKey(TLang *lng, TMacro *mc, struct EditorGetString& gs, TEditorPos& ep, char *line)
{
  char *ptr = lstrcpy(line, gs.StringText)+ep.Col;
  if ( !inComment(lng, line, ep.Col) )
  {
    TExpandType at = GetMacroPos(line, ptr, 0);
    if ( CheckMacroPos(mc, at) != NULL )
    {
      RunMacro(mc, lng);
      redraw();
      return 1;
    }
    if ( FindMacroKeyEx(lng, mc, at) )
    {
      RunMacro(mc, lng);
      redraw();
      return 1;
    }
  }
  return 0;
}

static int findLngID(char* fn)
{
  for ( unsigned i = 0 ; i < langColl.getCount() ; i++ )
  {
    TLang *lng = (TLang*)(langColl.at(i));
    if ( matched(lng->mask, fn) )
      return i;
  }
  return -1;
}

int WINAPI _export ProcessEditorEvent(int Event, void *Param)
{
  struct EditorInfo ei;
  initEList();
  switch ( Event )
  {
    case EE_READ:
      Info.EditorControl(ECTL_GETINFO, &ei);
      eList->insert(ei.EditorID, findLngID(ei.FileName));
      return 0;
    case EE_CLOSE:
      eList->removeID(*((int*)Param));
      return 0;
    default:
      return 0;
  }
}

int WINAPI _export ProcessEditorInput(const INPUT_RECORD *Rec)
{
  struct EditorInfo ei;
  if ( pluginStop || pluginBusy || Rec->EventType != KEY_EVENT || Rec->Event.KeyEvent.bKeyDown == 0 )
    return PROCESS_EVENT;
  Info.EditorControl(ECTL_GETINFO, &ei);
  initEList();
  if ( eList->findID(ei.EditorID) == -1 )
    eList->insert(ei.EditorID, findLngID(ei.FileName));
  int id = eList->findLang(ei.EditorID);
  if ( id != -1 )
  {
    TLang *lng = (TLang*)(langColl.at(id));
    if ( lng )
    {
      char line[1024];
      struct EditorGetString gs;
      TEditorPos epos = EditorGetPos();
      TMacro *fm;
      if ( ( fm = FindMacroKey(lng, Rec) ) != NULL )
      {
        EditorGetString(&gs);
        if ( checkMultipleChoice(lng, fm, expAnyWhere) )
          return TryMacroKey(lng, fm, gs, epos, line) ? IGNORE_EVENT : PROCESS_EVENT;
        return IGNORE_EVENT;
      }
      if ( Rec->Event.KeyEvent.dwControlKeyState & ( RIGHT_ALT_PRESSED | LEFT_ALT_PRESSED | RIGHT_CTRL_PRESSED | LEFT_CTRL_PRESSED ) )
        return PROCESS_EVENT;
      char c = Rec->Event.KeyEvent.uChar.AsciiChar;
      if ( c == '\t' || Rec->Event.KeyEvent.wVirtualKeyCode == VK_TAB )
      {
        EditorProcessTabKey(lng->TabSize, epos.Col, 1);
        return IGNORE_EVENT;
      }
      int inImm, spORret = ( c == ' ' || Rec->Event.KeyEvent.wVirtualKeyCode == VK_RETURN );
      if ( lng->ignoreCase )
        inImm = lstrchr(lng->imm, FSF.LLower(c)) || lstrchr(lng->imm, FSF.LUpper(c));
      else
        inImm = lstrchr(lng->imm, c) != NULL;
      if ( !inImm && epos.Col == 0 )
        return PROCESS_EVENT;
      if ( ! ( spORret || ( c && inImm ) ) )
        return PROCESS_EVENT;
      if ( c == ' ' )
      {
        EditorGetString(&gs);
        return TryMacro(lng, gs, epos, line) ? IGNORE_EVENT : PROCESS_EVENT;
      }
      else
      {
        int ret = PROCESS_EVENT;
        EditorGetString(&gs);
        if ( !inComment(lng, gs.StringText, epos.Col) )
        {
          if ( !spORret )
          {
            pluginBusy = 1;
            EditorProcessKey(c);
            pluginBusy = 0;
            ret = IGNORE_EVENT;
            EditorGetString(&gs);
          }
          char *str = FirstNonSpace(gs.StringText);
          if ( str == NULL )
            return ret;
          int spacesCount = str-gs.StringText;
          lstrcpy(line, FSF.Trim(str));
          if ( lng->ignoreCase )
          {
            StrLower(line);
            c = (char)FSF.LLower(c);
          }
          for ( unsigned j = 0 ; j < lng->indentColl.getCount() ; j++ )
          {
            char ic = ((TIndent*)(lng->indentColl.at(j)))->immChar;
            if ( lng->ignoreCase )
              ic = (char)FSF.LLower(ic);
            if ( spORret || ( !spORret && ( ic == c ) ) )
              if ( TryIndent(spORret, lng, j, line, str, spacesCount) )
                return IGNORE_EVENT;
          }
          return ret;
        }
      }
    }
  }
  return PROCESS_EVENT;
}

HANDLE WINAPI _export OpenPlugin(int, int)
{
  if ( !IsOldFar )
  {
    EditorInfo ei;
    Info.EditorControl(ECTL_GETINFO, &ei);
    initEList();
    int n = eList->findID(ei.EditorID);
    if ( n == -1 )
      n = eList->insert(ei.EditorID, findLngID(ei.FileName));
    if ( n != -1 )
    {
      TEInfo *te = (TEInfo*)eList->at(n);
      if ( te )
      {
        unsigned lc = langColl.getCount();
        FarMenuItem *amenu = new FarMenuItem[lc+2];
        if ( amenu )
        {
          for ( unsigned i = 0 ; i < lc ; i++ )
          {
            FSF.sprintf(amenu[i].Text, "%s", ((TLang*)langColl.at(i))->desc);
            amenu[i].Selected = (unsigned)(te->lang) == i;
            amenu[i].Checked = amenu[i].Separator = false;
          }
          amenu[lc].Selected = amenu[lc].Checked = false;
          amenu[lc].Separator = true;
          *amenu[lc].Text = 0;
          FSF.sprintf(amenu[lc+1].Text, "%s", GetMsg(MOtherLang));
          amenu[lc+1].Selected = te->lang == -1;
          amenu[lc+1].Checked = amenu[lc+1].Separator = false;
          int res = Info.Menu(Info.ModuleNumber,-1,-1,0,FMENU_WRAPMODE,GetMsg(MSelectLang),NULL,NULL,NULL,NULL,amenu,lc+2);
          if ( res != -1 )
          {
            if ( pluginStop )
              SetRegKey(HKEY_CURRENT_USER, "", "Disable", pluginStop = 0);
            te->lang = ( (unsigned)res < lc ) ? res : -1;
          }
          delete [] amenu;
        }
      }
    }
  }
  return INVALID_HANDLE_VALUE;
}
